@@ -48,9 +48,9 @@
 
 #define H2_CUSTOM_ERRORS
 #define H2EC1(U,v,g,r,d)	\
-	const struct h2_error_s H2CE_##U[1] = {{#U,d,v,0,1,g,r}};
+	const struct h2_error_s H2CE_##U[1] = {{"H2CE_" #U,d,v,0,1,g,r}};
 #define H2EC2(U,v,g,r,d)	\
-	const struct h2_error_s H2SE_##U[1] = {{#U,d,v,1,0,g,r}};
+	const struct h2_error_s H2SE_##U[1] = {{"H2SE_" #U,d,v,1,0,g,r}};
 #define H2EC3(U,v,g,r,d) H2EC1(U,v,g,r,d) H2EC2(U,v,g,r,d)
 #define H2_ERROR(NAME, val, sc, goaway, reason, desc)	\
 	H2EC##sc(NAME, val, goaway, reason, desc)
@@ -150,7 +150,7 @@ h2_new_req(struct h2_sess *h2, unsigned stream, struct req *req)
 
 	ASSERT_RXTHR(h2);
 	if (req == NULL)
-		req = Req_New(h2->sess);
+		req = Req_New(h2->sess, NULL);
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
 
 	r2 = WS_Alloc(req->ws, sizeof *r2);
@@ -192,6 +192,8 @@ h2_del_req(struct worker *wrk, struct h2_req *r2)
 	--h2->refcnt;
 	/* XXX: PRIORITY reshuffle */
 	VTAILQ_REMOVE(&h2->streams, r2, list);
+	if (r2->req == h2->new_req)
+		h2->new_req = NULL;
 	Lck_Unlock(&sp->mtx);
 
 	assert(!WS_IsReserved(r2->req->ws));
@@ -327,6 +329,7 @@ h2_rx_push_promise(struct worker *wrk, struct h2_sess *h2, struct h2_req *r2)
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	ASSERT_RXTHR(h2);
 	CHECK_OBJ_ORNULL(r2, H2_REQ_MAGIC);
+
 	// rfc7540,l,2262,2267
 	H2S_Lock_VSLb(h2, SLT_SessError, "H2: rx push promise");
 	return (H2CE_PROTOCOL_ERROR);
@@ -335,14 +338,14 @@ h2_rx_push_promise(struct worker *wrk, struct h2_sess *h2, struct h2_req *r2)
 /**********************************************************************
  */
 
-static h2_error
-h2_rapid_reset(struct worker *wrk, struct h2_sess *h2, struct h2_req *r2)
+int
+h2_rapid_reset_check(struct worker *wrk, struct h2_sess *h2,
+    const struct h2_req *r2)
 {
 	vtim_real now;
-	vtim_dur d;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
-	ASSERT_RXTHR(h2);
+	CHECK_OBJ_NOTNULL(h2, H2_SESS_MAGIC);
 	CHECK_OBJ_NOTNULL(r2, H2_REQ_MAGIC);
 
 	if (h2->rapid_reset_limit == 0)
@@ -354,6 +357,23 @@ h2_rapid_reset(struct worker *wrk, struct h2_sess *h2, struct h2_req *r2)
 	if (now - r2->req->t_first > h2->rapid_reset)
 		return (0);
 
+	return (1);
+}
+
+h2_error
+h2_rapid_reset_charge(struct worker *wrk, struct h2_sess *h2,
+    const struct h2_req *r2)
+{
+	vtim_real now;
+	vtim_dur d;
+	h2_error h2e = NULL;
+
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	AN(H2_SEND_HELD(h2, r2));
+	CHECK_OBJ_NOTNULL(r2, H2_REQ_MAGIC);
+
+	now = VTIM_real();
+
 	d = now - h2->last_rst;
 	h2->rst_budget += h2->rapid_reset_limit * d /
 	    h2->rapid_reset_period;
@@ -361,18 +381,21 @@ h2_rapid_reset(struct worker *wrk, struct h2_sess *h2, struct h2_req *r2)
 	    h2->rapid_reset_limit);
 	h2->last_rst = now;
 
-	if (h2->rst_budget < 1.0) {
-		H2S_Lock_VSLb(h2, SLT_SessError, "H2: Hit RST limit. Closing session.");
-		return (H2CE_RAPID_RESET);
-	}
 	h2->rst_budget -= 1.0;
-	return (0);
+
+	if (h2->rst_budget < 0) {
+		H2S_Lock_VSLb(h2, SLT_SessError, "H2: Hit RST limit. Closing session.");
+		h2e = H2CE_RAPID_RESET;
+		H2_Send_GOAWAY(wrk, h2, r2, h2e);
+	}
+
+	return (h2e);
 }
 
 static h2_error v_matchproto_(h2_rxframe_f)
 h2_rx_rst_stream(struct worker *wrk, struct h2_sess *h2, struct h2_req *r2)
 {
-	h2_error h2e;
+	h2_error h2e = NULL;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	ASSERT_RXTHR(h2);
@@ -384,7 +407,11 @@ h2_rx_rst_stream(struct worker *wrk, struct h2_sess *h2, struct h2_req *r2)
 	}
 	if (r2 == NULL)
 		return (0);
-	h2e = h2_rapid_reset(wrk, h2, r2);
+	if (h2_rapid_reset_check(wrk, h2, r2)) {
+		H2_Send_Get(wrk, h2, r2);
+		h2e = h2_rapid_reset_charge(wrk, h2, r2);
+		H2_Send_Rel(h2, r2);
+	}
 	h2_kill_req(wrk, h2, r2, h2_streamerror(vbe32dec(h2->rxf_data)));
 	return (h2e);
 }
@@ -411,19 +438,14 @@ h2_rx_goaway(struct worker *wrk, struct h2_sess *h2, struct h2_req *r2)
 static void
 h2_tx_goaway(struct worker *wrk, struct h2_sess *h2, h2_error h2e)
 {
-	char b[8];
-
 	ASSERT_RXTHR(h2);
 	AN(h2e);
 
 	if (h2->goaway || !h2e->send_goaway)
 		return;
 
-	h2->goaway = 1;
-	vbe32enc(b, h2->highest_stream);
-	vbe32enc(b + 4, h2e->val);
 	H2_Send_Get(wrk, h2, h2->req0);
-	H2_Send_Frame(wrk, h2, H2_F_GOAWAY, 0, 8, 0, b);
+	H2_Send_GOAWAY(wrk, h2, h2->req0, h2e);
 	H2_Send_Rel(h2, h2->req0);
 }
 
@@ -462,6 +484,8 @@ h2_rx_window_update(struct worker *wrk, struct h2_sess *h2, struct h2_req *r2)
 
 /**********************************************************************
  * Incoming PRIORITY, possibly an ACK of one we sent.
+ *
+ * deprecated, rfc9113,l,1103,1104
  */
 
 static h2_error v_matchproto_(h2_rxframe_f)
@@ -650,6 +674,7 @@ h2_end_headers(struct worker *wrk, struct h2_sess *h2,
 		h2_del_req(wrk, r2);
 		return (h2e);
 	}
+	req->t_req = VTIM_real();
 	VSLb_ts_req(req, "Req", req->t_req);
 
 	// XXX: Smarter to do this already at HPACK time into tail end of
@@ -733,31 +758,30 @@ h2_rx_headers(struct worker *wrk, struct h2_sess *h2, struct h2_req *r2)
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	ASSERT_RXTHR(h2);
 
-	if (r2 == NULL) {
-		if (h2->rxf_stream <= h2->highest_stream) {
-			H2S_Lock_VSLb(h2, SLT_SessError, "H2: new stream ID < highest stream");
-			return (H2CE_PROTOCOL_ERROR);	// rfc7540,l,1153,1158
-		}
-		/* NB: we don't need to guard the read of h2->open_streams
-		 * because headers are handled sequentially so it cannot
-		 * increase under our feet.
-		 */
-		if (h2->open_streams >=
-		    (int)h2->local_settings.max_concurrent_streams) {
-			H2S_Lock_VSLb(h2, SLT_Debug,
-			     "H2: stream %u: Hit maximum number of "
-			     "concurrent streams", h2->rxf_stream);
-			return (H2SE_REFUSED_STREAM);	// rfc7540,l,1200,1205
-		}
-		h2->highest_stream = h2->rxf_stream;
-		r2 = h2_new_req(h2, h2->rxf_stream, NULL);
-	}
-	CHECK_OBJ_NOTNULL(r2, H2_REQ_MAGIC);
-
-	if (r2->state != H2_S_IDLE) {
+	if (r2 != NULL) {
 		H2S_Lock_VSLb(h2, SLT_SessError, "H2: rx headers on non-idle stream");
-		return (H2CE_PROTOCOL_ERROR);	// XXX spec ?
+		return (H2CE_PROTOCOL_ERROR);	// rfc9113,l,887,891
 	}
+
+	if (h2->rxf_stream <= h2->highest_stream) {
+		H2S_Lock_VSLb(h2, SLT_SessError, "H2: new stream ID < highest stream");
+		return (H2CE_PROTOCOL_ERROR);	// rfc7540,l,1153,1158
+	}
+        /* NB: we don't need to guard the read of h2->open_streams
+         * because headers are handled sequentially so it cannot
+         * increase under our feet.
+         */
+        if (h2->open_streams >=
+	    (int)h2->local_settings.max_concurrent_streams) {
+		H2S_Lock_VSLb(h2, SLT_Debug,
+		    "H2: stream %u: Hit maximum number of "
+		    "concurrent streams", h2->rxf_stream);
+		return (H2SE_REFUSED_STREAM);	// rfc7540,l,1200,1205
+	}
+	h2->highest_stream = h2->rxf_stream;
+	r2 = h2_new_req(h2, h2->rxf_stream, NULL);
+	CHECK_OBJ_NOTNULL(r2, H2_REQ_MAGIC);
+	assert(r2->state == H2_S_IDLE);
 	r2->state = H2_S_OPEN;
 
 	req = r2->req;
@@ -771,8 +795,7 @@ h2_rx_headers(struct worker *wrk, struct h2_sess *h2, struct h2_req *r2)
 	req->sp = h2->sess;
 	req->transport = &HTTP2_transport;
 
-	req->t_first = VTIM_real();
-	req->t_req = VTIM_real();
+	req->t_first = h2->t1;
 	req->t_prev = req->t_first;
 	VSLb_ts_req(req, "Start", req->t_first);
 	req->acct.req_hdrbytes += h2->rxf_len;
@@ -1415,8 +1438,8 @@ h2_sweep(struct worker *wrk, struct h2_sess *h2)
 		}
 		switch (r2->state) {
 		case H2_S_CLOSED:
-			if (!r2->scheduled)
-				h2_del_req(wrk, r2);
+			AZ(r2->scheduled);
+			h2_del_req(wrk, r2);
 			break;
 		case H2_S_CLOS_REM:
 			if (!r2->scheduled) {
@@ -1447,6 +1470,27 @@ h2_sweep(struct worker *wrk, struct h2_sess *h2)
 	return (h2e);
 }
 
+/*
+ * if we have received end_headers, the new request is started
+ * if we have not received end_stream, DATA frames are expected later
+ *
+ * neither of these make much sense to output here
+ *
+ * goaway currently is always 0, see #4285
+ */
+static void
+h2_htc_debug(enum htc_status_e hs, struct h2_sess *h2)
+{
+	const char *s, *r;
+
+	if (LIKELY(VSL_tag_is_masked(SLT_Debug)))
+		return;
+
+	HTC_Status(hs, &s, &r);
+	H2S_Lock_VSLb(h2, SLT_Debug, "H2: HTC %s (%s) frame=%s goaway=%d",
+	    s, r, h2->htc->rxbuf_b == h2->htc->rxbuf_e ? "complete" : "partial",
+	    h2->goaway);
+}
 
 /***********************************************************************
  * Called in loop from h2_new_session()
@@ -1476,27 +1520,34 @@ h2_rxframe(struct worker *wrk, struct h2_sess *h2)
 	if (h2->goaway && h2->open_streams == 0)
 		return (0);
 
+	h2->t1 = NAN;
 	VTCP_blocking(*h2->htc->rfd);
-	hs = HTC_RxStuff(h2->htc, h2_frame_complete, NULL, NULL, NAN,
+	hs = HTC_RxStuff(h2->htc, h2_frame_complete, &h2->t1, NULL, NAN,
 	    VTIM_real() + 0.5, NAN, h2->local_settings.max_frame_size + 9);
 
 	h2e = NULL;
 	switch (hs) {
+	case HTC_S_EOF:
+		h2_htc_debug(hs, h2);
+		h2e = H2CE_NO_ERROR;
+		break;
 	case HTC_S_COMPLETE:
 		h2->sess->t_idle = VTIM_real();
 		if (h2->do_sweep)
 			h2e = h2_sweep(wrk, h2);
 		break;
 	case HTC_S_TIMEOUT:
+		//// #4279
+		// h2_htc_debug(hs, h2);
 		h2e = h2_sweep(wrk, h2);
 		break;
 	default:
+		HTC_Status(hs, &s, &r);
+		H2S_Lock_VSLb(h2, SLT_SessError, "H2: HTC %s (%s)", s, r);
 		h2e = H2CE_ENHANCE_YOUR_CALM;
 	}
 
 	if (h2e != NULL && h2e->connection) {
-		HTC_Status(hs, &s, &r);
-		H2S_Lock_VSLb(h2, SLT_SessError, "H2: HTC %s (%s)", s, r);
 		h2->error = h2e;
 		h2_tx_goaway(wrk, h2, h2e);
 		return (0);
